/* *******************************************************************
   Board Settings for
   ESP-12F DC 5V/12V/24V 16 Channel Relay Board
   ESP12F_Relay_X16 with two shift registers
   Controller:   ESP8266
   Setting:      NodeMCU 1.0 (ESP-12E Module)
                 4MB (FS:2MB OTA: ~1019KB)
   Description:  https://forum.arduino.cc/t/sketch-fur-esp8266-16-relay/1123972/

   Amazon:       https://amzn.to/42z1PIp
   Aliexpress:   https://s.click.aliexpress.com/e/_DktPvBT

  ********************************************************************/

#pragma once
#if USE_BOARD == 72                    // ESP8266: example board: 
#define TXT_BOARDNAME "ESP12F Relay X16"       // the name of the board
HardwareSerial &Terminal = Serial;

const uint8_t flashPin = 0;            // GPIO00/D3 on NodeMCU the flash Button - this is for testing an input
const uint8_t inputAPin = 15;          // GPIO15/D8 on NodeMCU
const uint8_t stcpPin = 12;            // GPIO12    74x595 RCLK/STCP
const uint8_t shcpPin = 13;            // GPIO13    74x595 SRCLK/SHCP
const uint8_t serPin = 14;             // GPIO14    74x595 SER/DS
const uint8_t oePin = 5;               // GPIO05/D1 74x595 OE/output enable active low

// assign pins to arrays:
Pin inputPin[] {
  // the pin    a name for the website
  {inputAPin,   "External"},
  {flashPin,    "Flash Jumper"},
};
// or
//#define NO_INPUTPIN

OutputPin outputPin[] {
  // the pin     a name  LOW active (or high active)
  {2,            "IO02", LOW},     // on some ESP8266 there is a blue LED on the ESP (low active)
  {16,           "IO16", LOW},     // on a NodeMCU there is a LED (low active)
};
// or
//#define NO_OUTPUTPIN

// this board has no wired PWM output so we need this precompiler define
#define NO_PWMPIN

const byte noOfShiftIc = 2; // number of shift registers

// if you don't need all 16 outputs you can delete
OutputPin shiftPin[] {
  // the pin  a name     LOW active (or high active)
  {0,         "Shift A", HIGH},
  {1,         "Shift B", HIGH},
  {2,         "Shift C", HIGH},
  {3,         "Shift D", HIGH},
  {4,         "Shift E", HIGH},
  {5,         "Shift F", HIGH},
  {6,         "Shift G", HIGH},
  {7,         "Shift H", HIGH},
  {8,         "Shift I", HIGH},
  {9,         "Shift J", HIGH},
  {10,        "Shift K", HIGH},
  {11,        "Shift L", HIGH},
  {12,        "Shift M", HIGH},
  {13,        "Shift O", HIGH},
  {14,        "Shift P", HIGH},
  {15,        "Shift Q", HIGH},
};

const uint8_t wifiActivityPin = 255;  // set to 255 if activity LED is not needed
#endif
