/* *******************************************************************
   Webserver

   How it all works together:

   page   0.htm         includes JavaScript j.js and f.css
          1.htm         just another page
   script j.js          the JavaScript requests the JSON
          json          returns data as JSON
          cmd           a resource doing actions and not returning content (just the http header 204 no content)
   css    f.css         a css stylesheet for all on flash (program) memory
   ***************************************************************** */

void handleNotFound() {
  // Output a "404 not found" page. It includes the parameters which comes handy for test purposes.
  Serial.println(F("D015 handleNotFound()"));
  String message;
  message += F("404 - File Not Found\n"
               "URI: ");
  message += server.uri();
  message += F("\nMethod: ");
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += F("\nArguments: ");
  message += server.args();
  message += F("\n");
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}

void handle204() {
  server.send(204);                // this page doesn't send back content
}

// the top of the page till tag <main>
void addTop(String &message) {
  message =  F("<!DOCTYPE html>\n"
               "<html lang='en'>\n"
               "<head>\n"
               "<title>" TXT_BOARDNAME " - Board " );
  message += USE_BOARD;
  message += F("</title>\n"
               "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\">\n"
               "<meta name=\"viewport\" content=\"width=device-width\">\n"
               "<link rel='stylesheet' type='text/css' href='f.css'>\n"
               "<script src='j.js'></script>\n"
               "</head>\n");
  message += F("<body>\n");
  message += F("<header>\n<h1>" TXT_BOARDNAME " - Board " );
  message += USE_BOARD;
  message += F("</h1>\n"
               "<nav><p><a href=\"/\">[Home]</a> <a href=\"1.htm\">[More]</a> <a href=\"9.htm\">[Debug]</a> </p></nav>\n</header>\n"
               "<main>\n");
}

// the bottom of the page from tag </main> downwards
void addBottom(String &message) {
  message += F("</main>\n"
               "<footer>\n<p>");                 // The footer will display the uptime, the version of the sketch and the compile date/time
  if (ss > 604800) {
    message += F("<span id='week'>");
    message +=  ((ss / 604800) % 52);
    message += F("</span> weeks ");
  }
  if (ss > 86400) {
    message += F("<span id='day'>");
    message += ((ss / 86400) % 7);
    message += F("</span> days ");
  }
  if (ss > 3600) {
    message += F("<span id='hour'>");
    message += ((ss / 3600) % 24);
    message += F("</span> hours ");
  }
  message += F("<span id='min'>");
  message += ((ss / 60) % 60);
  message += F("</span> minutes ");
  message += F("<span id='sec'>");
  message += (ss % 60);
  message += F("</span> seconds since startup | Version " VERSION );
  message += F(" | <span style='white-space:nowrap'>" __DATE__ " " __TIME__ "</span>"
               " <br><span id='msg'></span>"              // just an empty space for some debug information, hearbeat etc
               "</p>\n</footer>\n</body>\n</html>");
  server.send(200, "text/html", message);
}

// the html output
// finally check your output if it is valid html: https://validator.w3.org
// *** HOME ***  0.htm
void handlePage() {
  String message;
  addTop(message);

#ifndef NO_PWMPIN
  /* Sliders for PWM pins */
  if (sizeof(pwmPin)) {                                      // only if pwmPins are defined...
    message += F("<article>\n"
                 "<h2>PWM</h2>\n"                            // example how to switch/toggle an output
                 "<p>Slide to change output.</p>\n");
    for (size_t i = 0; i < noOfPwmPin; i++) {
      message += F("<p style='padding-left:10px'>");                         // MISSING - that's not nice ... to have a style in source...
      message += F("<input type='range' step='1' min='0' max='63' value='"); // the slider has only 0..63 steps, but the function will send 0..1023
      message += pwmPin[i].value;
      message += F("' id='valuepwmPin"); // we need a unique ID to be able to assign values via JavaScript
      message += i;
      message += F("' onInput='firesmooth(this.id, this.value)'><span class='sliderlabel'>");
      message += pwmPin[i].label;
      message += F(" (");
      message += pwmPin[i].pin;
      message += F(")</span></p>\n");
    }
    message += F("</article>\n");
  }
#endif
#ifndef NO_OUTPUTPIN
  /* display nice ON/OFF rocket switches instead of checkboxes */
  message += F("<article>\n"
               "<h2>Switch</h2>\n"                         // example how to switch/toggle an output
               "<p>Toggle switches for on/off outputs</p>\n");
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++) {
    message += F("<div class='toggle'><label><input type='checkbox' ");
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("value='checked' checked='checked' ");
    message += F("onclick='firecb(this.id)' id='cboutputPin");
    message += i;
    message += F("'><span class='slider'></span><span class='desc'>");
    message += outputPin[i].label;
    message += F(" (GPIO");
    message += outputPin[i].pin;
    message += F(")</span></label></div>\n");
  }
  message += F("</article>\n");
#endif
#ifndef NO_SHIFTPIN
  /* display nice ON/OFF rocket switches instead of checkboxes */
  message += F("<article>\n"
               "<h2>Switch Shift Register</h2>\n"                         // example how to switch/toggle an output
               "<p>Toggle switches for outputs on shift registers</p>\n");
  for (size_t i = 0; i < sizeof(shiftPin) / sizeof(shiftPin[0]); i++) {
    message += F("<div class='toggle'><label><input type='checkbox' ");
    if (shiftRead(shiftPin[i].pin) == shiftPin[i].ON)
      message += F("value='checked' checked='checked' ");
    message += F("onclick='firecb(this.id)' id='cbshiftPin");
    message += i;
    message += F("'><span class='slider'></span><span class='desc'>");
    message += shiftPin[i].label;
    message += F(" (Q");
    message += shiftPin[i].pin % 8;
    message += F(")</span></label></div>\n");
  }
  message += F("</article>\n");
#endif
#if defined(NO_OUTPUTPIN) && defined(NO_INPUTPIN) && defined(NO_PWMPIN) && defined(NO_SHIFTPIN)
  message += F("<p>You should define outputPin, inputPin, pwmPin and/or shiftPin</p>\n");
#endif
  message += F("<iframe name='i' style='display:none'></iframe>\n");         // hack to keep the href button press in the window

  addBottom(message);
  server.send(200, "text/html", message);
}

// *** Other ***  1.htm
// just some basic examples
void handlePage1() {
  String message;
  addTop(message);

#ifndef NO_OUTPUTPIN
  /* Simple HREF links styled as buttons to toggle the outputs */
  message += F("<article>\n"
               "<h2>Switch</h2>\n"                                             // example how to switch/toggle an output
               "<p>Example how to switch/toggle outputs, or to initiate actions. The buttons are 'fake' buttons and only styled by CSS. Click to toggle the output.</p>\n");
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++) {
    message += F("<p><a class='btn size2' href='cmd?toggle=");
    message += i;
    message += F("' target='i'>&nbsp;");
    message += outputPin[i].label;
    message += F("&nbsp;</a></p>\n");
  }
  message += F("</article>\n");
#endif

#ifndef NO_PWMPIN
  /* Simple HREF links styled as buttons to toggle the PWM outputs */
  if (sizeof(pwmPin)) {
    message += F("<article>\n"
                 "<h2>PWM Pins</h2>\n"                                           // example how to switch/toggle an output
                 "<p>Click to toggle the PWM output (full on / full off)</p>\n");
    for (size_t i = 0; i < noOfPwmPin; i++) {
      message += F("<p><a class='btn size2' href='cmd?pwmToggle=");
      message += i;
      message += F("' target='i'>&nbsp;");
      message += pwmPin[i].label;
      message += F("&nbsp;</a></p>\n");
    }
    message += F("</article>\n");
  }
#endif

#ifndef NO_SHIFTPIN
  /* Simple HREF links styled as buttons to toggle the outputs */
  message += F("<article>\n"
               "<h2>Switch Shift</h2>\n"                                             // example how to switch/toggle an output
               "<p>Example how to switch/toggle shift registers. The buttons are 'fake' buttons and only styled by CSS. Click to toggle the output.</p>\n");
  for (size_t i = 0; i < sizeof(shiftPin) / sizeof(shiftPin[0]); i++) {
    message += F("<p><a class='btn size2' href='cmd?shiftToggle=");
    message += i;
    message += F("' target='i'>&nbsp;");
    message += shiftPin[i].label;
    message += F("&nbsp;</a></p>\n");
  }
  message += F("</article>\n");
#endif

#if defined(NO_OUTPUTPIN) && defined(NO_INPUTPIN) && defined(NO_PWMPIN) && defined(NO_SHIFTPIN)
  message += F("<p>You should define outputPin, inputPin, pwmPin and/or shiftPin</p>\n");
#endif
  message += F("<iframe name='i' style='display:none'></iframe>\n");         // hack to keep the href button press in the window

  addBottom(message);
  server.send(200, "text/html", message);
}

// *** Debug ***  9.htm
// just some value output
void handlePage9() {
  String message;
  addTop(message);

  /* output of current values */

#ifdef ARDUINO_ARCH_ESP8266
  message += F("<article>\n"
               "<h2>Values (with update)</h2>\n");
  message += F("<p>Internal Voltage measured by ESP: <span id='internalVcc'>");  // example how to show values on the webserver
  message += ESP.getVcc();  // only available on ESP8266
  message += F("</span>mV</p>\n"
               "</article>\n");
#endif

#ifndef NO_INPUTPIN
  message += F("<article>\n"
               "<h2>Input Pins</h2>\n");
  for (size_t i = 0; i < sizeof(inputPin) / sizeof(inputPin[0]); i++) {
    message += F("<p>inputPin");
    message += i;
    message += F(" - ");
    message += inputPin[i].label;
    message += F(": <span id='inputPin");
    message += i;
    message += F("'>");
    message += digitalRead(inputPin[i].pin);
    message += F("</span> ");
    message += F("</p>\n");
  }
  message += F("</article>\n");
#endif

#ifndef NO_PWMPIN
  message += F("<article>\n"
               "<h2>PWM Pins</h2>\n");
  for (size_t i = 0; i < noOfPwmPin; i++) {
    message += F("<p>pwmPin");
    message += i;
    message += F(" - ");
    message += pwmPin[i].label;
    message += F(": <span id='pwmPin");
    message += i;
    message += F("'>");
    message += pwmPin[i].value;
    message += F("</span>");
    message += F("</p>\n");
  }
  message += F("</article>\n");
#endif
#ifndef NO_OUTPUTPIN
  message += F("<article>\n"
               "<h2>Output Pins</h2>\n");
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++) {
    message += F("<p>outputPin");
    message += i;
    message += F(" - ");
    message += outputPin[i].label;
    message += F(" (GPIO");
    message += outputPin[i].pin;
    message += F("): <span id='outputPin");
    message += i;
    message += F("'>");
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("1");
    else
      message += F("0");
    message += F("</span>");
    message += F("</p>\n");
  }
  message += F("</article>\n");
#endif

#ifndef NO_SHIFTPIN
  message += F("<article>\n"
               "<h2>Shift Registers</h2>\n");
  for (size_t i = 0; i < sizeof(shiftPin) / sizeof(shiftPin[0]); i++) {
    message += F("<p>shiftPin");
    message += i;
    message += F(" - ");
    message += shiftPin[i].label;
    message += F(" (Q");
    message += shiftPin[i].pin % 8;
    message += F("): <span id='shiftPin");
    message += i;
    message += F("'>");
    if (shiftRead(shiftPin[i].pin) == shiftPin[i].ON)
      message += F("1");
    else
      message += F("0");
    message += F("</span>");
    message += F("</p>\n");
  }
  message += F("</article>\n");
#endif

#if defined(NO_OUTPUTPIN) && defined(NO_INPUTPIN) && defined(NO_PWMPIN) && defined(NO_SHIFTPIN)
  message += F("<article>\n"
               "<p>You should define outputPin, inputPin, pwmPin and/or shiftPin</p>\n"
               "</article>\n");
#endif

  addBottom(message);
  server.send(200, "text/html", message);
}

void handleCss() {
  // output of stylesheet
  // this is a straight forward example how to generat a static page from program memory
  String message;
  message = F(":root{\n"                          // some values (e.g. colors) might be used several times in several classes, therefore as variables
              "--no-col-main: #1279b1;\n"         // color main theme, e.g. h1 background
              "--no-col-on: #28A745;\n"           // color for on, active
              "--no-col-accent: orange;\n"
              "--no-col-warning: #ffc108;\n"      // color for warnings "yellow"
              "--no-col-danger: #DC3445;\n"       // color for danger, errors, "red"
              "--no-col-bg: white;\n"             // color main background
              "--no-col-bg0: #00487A;\n"          // color accents from light to darker
              "--no-col-bg1: #D0D0D0;\n"
              "--no-col-bg2: #F5F5F5;\n"
              "--no-col-bg5: silver;\n"           // e.g.: footer background
              "--no-col-bgm: #263238;\n"          // darkest background, e.g. for modal windows
              "--no-col-font: white;\n"           // contrast to main color, e.g. in h1
              "--no-col-font5: dimgrey;\n"        // e.g.: footer text
              "}\n"
              "*{font-family:sans-serif}\n"
              "body{margin:10px;background:var(--no-col-bg)}\n"
              "h1, h2{color:white;background:var(--no-col-main);text-align:center}\n"
              "h1{margin:1px;padding:5px;font-size:1.2em}\n"
              "h2{font-size:1.0em}\n"
              "h3{font-size:0.9em}\n"
              "a{text-decoration:none;color:dimgray;text-align:center}\n"
              "main{text-align:center}\n"
              "article{vertical-align:top;display:inline-block;margin:0.2em;padding:0.1em;border-style:solid;border-width:1px;border-color:var(--no-col-bg1);background-color:var(--no-col-bg2);width:20em;text-align:left}\n"
              "article h2{margin-top:0;padding-bottom:1px}\n"
              "section{margin-bottom:0.2em;clear:both}\n"
              "table{border-collapse:separate;border-spacing:0 0.2em}\n"
              "th, td{background-color:var(--no-col-bg1)}\n"
              "footer p{margin-bottom:5px;color:var(--no-col-font5);background:var(--no-col-bg5);font-size:0.7em;text-align:center}\n"
              "nav{background-color:var(--no-col-bg5);margin:1px;padding:5px;font-size:0.8em}\n"
              "nav a{color:var(--no-col-font5);padding:10px;text-decoration:none}\n"
              "nav a:hover{text-decoration:underline}\n"
              "nav p{margin:0px;padding:0px}\n"
              "button, input[type='button'], input[type='submit'], .btn, .on, .off{margin-top:0;margin-bottom:0.2em;margin-left:4em;font-size:1.4em;background:var(--no-col-main);border-style:solid;border-width:1px;border-radius:.3rem;border-style:outset;width:5em;height:1.5em;color:var(--no-col-font);text-decoration:none;text-align:center}\n"
              ".size2{display:flex;width:7em;height:1.5em;justify-content:center;align-items:center}\n"
              ".on{background:var(--no-col-on)}\n"
              ".off{background:var(--no-col-danger)}\n"
              // checkbox toggle
              ".toggle label {position:relative;display:inline-block;width:8em;height:3.5em}\n"
              ".toggle input {display:none}\n"
              ".toggle .slider {position:absolute;cursor:pointer;top:1.5em;left:2em;width:4em;height:2em;background-color:darkgrey;border-radius:1em;transition:all .3s ease-in-out}\n"
              ".toggle .slider::before {position:absolute;content:\"\";height:1.6em;width:1.6em;left:0.2em;bottom:0.2em;background-color:white;border-radius:50%;transition:all .3s ease-in-out}\n"
              ".toggle input:checked + .slider {background-color:var(--no-col-on)}\n" // some kind of green/on
              ".toggle input:checked + .slider::before {transform:translateX(1.9em)}\n"
              ".cb input[type=checkbox]{visibility: hidden}\n"
              ".cb{width:25px;height:25px;background:#ddd;margin:10px 5px;position:relative;box-shadow:0px 1px 3px rgba(0,0,0,0.5);display:inline-block}\n"
              ".cb label{display:block;width:20px;height:18px;cursor:pointer;position:absolute;top:2px;left:2px;z-index:2;background:darkgrey;box-shadow:inset 0px 1px 3px rgba(0,0,0,0.5);color:white}\n"
              ".cb input[type=checkbox]:checked + label{background:#5a9900}\n"
              ".desc {width:200px;height:25px;margin:30px 0px 0px 120px;position:relative;display:inline-block}\n" // the description beside the switch
              // slider
              "input[type=range]{-webkit-appearance:none;margin: 12px 0px;/*width: 100%;*/}\n"
              "input[type=range]:focus{outline:none}\n"
              "input[type=range]::-webkit-slider-runnable-track{width:100%;height:8.4px;cursor:pointer;animate:0.2s;box-shadow: 1px 1px 1px #000000,0px 0px 1px #0d0d0d;background:dark;border-radius:1.3px;border: 0.2px solid #010101}\n"
              "input[type=range]::-webkit-slider-thumb {box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border:1px solid #000000;height:36px;width:16px;border-radius:3px;background:#F5F5F5;cursor:pointer;-webkit-appearance:none;margin-top: -14px}\n"
              "input[type=range]::-moz-range-track{width:100%;height:8.4px;cursor:pointer;animate:0.2s;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;background: dark;border-radius: 1.3px;border: 0.2px solid #010101}\n"
              "input[type=range]::-moz-range-thumb{box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border:1px solid #000000;height:36px;width:16px;border-radius:3px;background:#F5F5F5;cursor: pointer}\n"
              "input[type=range]::-ms-track{width: 100%;height: 8.4px;cursor: pointer;animate: 0.2s;background:transparent;border-color:transparent;border-width:16px 0;color:transparent}\n"
              "input[type=range]::-ms-fill-lower{background:darkgray;border: 0.2px solid #010101;border-radius: 2.6px;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d}\n"
              "input[type=range]::-ms-fill-upper{background:dark;border:0.2px solid #010101;border-radius: 2.6px;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d}\n"
              "input[type=range]::-ms-thumb {box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border: 1px solid #000000;height:36px;width:16px;background:#F5F5F5;cursor:pointer;border-radius: 3px}\n"
              ".sliderlabel{vertical-align:top;margin:13px 0 0 20px;position:relative;display:inline-block}\n"
             );
  server.send(200, "text/css", message);
}

// Output: send data to browser as JSON
// after amodification always check if JSON is still valid. Just call the JSON (json) in your webbrowser and check.
// Terminal.println(F("D359 requested json"));
void handleJson() {
  String message = "";
  message = (F(" {\"ss\":"));          // Start of JSON and the first object "ss":
  message += millis() / 1000;
#ifdef ARDUINO_ARCH_ESP8266            // The ESP8266 can meassure the internal voltage
  message += (F(",\"internalVcc\":"));
  message += ESP.getVcc();
#endif
#ifndef NO_INPUTPIN
  for (size_t i = 0; i < sizeof(inputPin) / sizeof(inputPin[0]); i++) {
    message += (F(",\"inputPin"));
    message += i;
    message += (F("\":"));
    message += digitalRead(inputPin[i].pin);
  }
#endif
#ifndef NO_OUTPUTPIN
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++) {
    message += (F(",\"outputPin"));
    message += i;
    message += (F("\":"));
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("1");
    else
      message += F("0");
  }
#endif
#ifndef NO_PWMPIN
  for (size_t i = 0; i < noOfPwmPin; i++) {
    message += (F(",\"pwmPin"));
    message += i;
    message += (F("\":"));
    message += pwmPin[i].value;
  }
#endif
#ifndef NO_SHIFTPIN
  for (size_t i = 0; i < sizeof(shiftPin) / sizeof(shiftPin[0]); i++) {
    message += (F(",\"shiftPin"));
    message += i;
    message += (F("\":"));
    if (shiftRead(shiftPin[i].pin) == shiftPin[i].ON)
      message += F("1");
    else
      message += F("0");
  }
#endif
  message += (F("}"));                           // End of JSON
  server.send(200, "application/json", message); // set MIME type https://www.ietf.org/rfc/rfc4627.txt
}


void handleCommand() {
  // receive command and handle accordingly
  // Terminal.println(F("D429 handleCommand"));
  for (uint8_t i = 0; i < server.args(); i++) {
    Terminal.print(server.argName(i));
    Terminal.print(F(": "));
    Terminal.println(server.arg(i));
  }
  if (false) ;                                        // needed just in case the board would not have any output pins
#ifndef NO_OUTPUTPIN
  else if (server.argName(0) == "toggle") {                // the parameter which was sent to this server
    byte i = server.arg(0).toInt();
    if (i < sizeof(outputPin) / sizeof(outputPin[0])) {
      if (digitalRead(outputPin[i].pin) != outputPin[i].ON) {        // toggle: if the pin was on - switch it off and vice versa
        digitalWrite(outputPin[i].pin, outputPin[i].ON);
      } else {
        digitalWrite(outputPin[i].pin, !outputPin[i].ON);
      }
    }
    else {
      Terminal.println(F("E4447: invalid pin"));
    }
  }
  else if (server.argName(0) == "cb" && server.argName(1) == "v" && server.arg(0).substring(0, 11) == "cboutputPin") {  // a check base value from a slider
    byte i = server.arg(0).substring(11).toInt();          // we get a "cboutputPin"
    //Terminal.print("D453 cb i:"); Terminal.println(i);
    if (i < sizeof(outputPin) / sizeof(outputPin[0])) {
      if (server.arg(1) != "0")                            // we get a string not an integer
        digitalWrite(outputPin[i].pin, outputPin[i].ON);
      else
        digitalWrite(outputPin[i].pin, !outputPin[i].ON);
    }
    else {
      Terminal.println(F("E460: invalid pin"));
    }
  }
#endif

#ifndef NO_PWMPIN
  else if (server.argName(0) == "pwmToggle")     // the parameter which was sent to this server
  {
    byte i = server.arg(0).toInt();
    if (i < noOfPwmPin)
    {
#ifdef ARDUINO_ARCH_ESP32
      if (pwmPin[i].value < 512) {   // toggle: if the pin was on - switch it off and vice versa
        pwmPin[i].value = 1023;
      } else {
        pwmPin[i].value = 0;
      }
      ledcWrite(i, pwmPin[i].value); // write the channel (not the pin!)
#else
      if (digitalRead(pwmPin[i].pin) == LOW) {   // toggle: if the pin was on - switch it off and vice versa
        digitalWrite(pwmPin[i].pin, HIGH);       // MISSING tbc if rework similar to ESP32
        pwmPin[i].value = 1023;
      } else {
        digitalWrite(pwmPin[i].pin, LOW);
        pwmPin[i].value = 0;
      }
#endif
    }
    else {
      Terminal.println(F("E489: invalid pin"));
    }
  }
  else if (server.argName(0) == "range" && server.argName(1) == "v") {// a range based value from a slider
    byte i = server.arg(0).substring(11).toInt();    // we get a "valuepwmPin"
    if (i < noOfPwmPin) {
      //Terminal.print(F("D495 range i:")); Terminal.println(i);
      pwmPin[i].value = constrain(server.arg(1).toInt(), 0, 1023);
#ifdef ARDUINO_ARCH_ESP32
      ledcWrite(i, pwmPin[i].value); // write the channel (not the pin!)
#else
      analogWrite(pwmPin[i].pin, pwmPin[i].value);
#endif
    }
    else {
      Terminal.println(F("E504: invalid pin"));
    }
  }
#endif // von #ifndef NO_PWMPIN

#ifndef NO_SHIFTPIN
  else if (server.argName(0) == "shiftToggle") {           // the parameter which was sent to this server
    byte i = server.arg(0).toInt();
    if (i < sizeof(shiftPin) / sizeof(shiftPin[0])) {
      if (shiftRead(shiftPin[i].pin) != shiftPin[i].ON) {  // toggle: if the pin was on - switch it off and vice versa
        shiftWrite(shiftPin[i].pin, shiftPin[i].ON);
      } else {
        shiftWrite(shiftPin[i].pin, !shiftPin[i].ON);
      }
    }
    else {
      Terminal.println(F("E520: invalid pin"));
    }
  }
  else if (server.argName(0) == "cb" && server.argName(1) == "v" && server.arg(0).substring(0, 10) == "cbshiftPin") {  // a check base value from a slider
    byte i = server.arg(0).substring(10).toInt();          // we get a "cbshiftPin"
    //Terminal.print(F("D525 cb i:")); Terminal.println(i);
    if (i < sizeof(shiftPin) / sizeof(shiftPin[0])) {
      if (server.arg(1) != "0")                            // we get a string not an integer
        shiftWrite(shiftPin[i].pin, shiftPin[i].ON);
      else
        shiftWrite(shiftPin[i].pin, !shiftPin[i].ON);
    }
    else {
      Terminal.println(F("E533: invalid pin"));
    }
  }
#endif  // of #ifndef NO_SHIFTPIN

  else if (server.argName(0) == "CMD" && server.arg(0) == "RESET")   // Example how to reset the module. Just send ?CMD=RESET
  {
    Terminal.println(F("D540 will reset"));
    ESP.restart();
  }

  server.send(204);                      // this page doesn't send back content --> 204
}


void handleJs() {
  // Output: a fetch API / JavaScript
  // a function in the JavaScript uses fetch API to request a JSON file from the webserver and updates the values on the page if the object names and ID are the same
  String message;
  message += F("const url ='json';\n"
               "function renew(){\n"
               " document.getElementById('sec').style.color = 'blue';\n"                                            // if the timer starts the request, the second gets blue
               " if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&orarr;';\n"         // show activity: "retrive request"
               " fetch(url)\n" // Call the fetch function passing the url of the API as a parameter
               " .then(response => {return response.json();})\n"
               " .then(jo => {\n"
               "   document.getElementById('sec').innerHTML = Math.floor(jo['ss'] % 60);\n"                         // example how to change a value in the HTML page
               "   for (var i in jo)\n"
               "   {\n"
               "    if (document.getElementById(i)) document.getElementById(i).innerHTML = jo[i];\n"                // as long as the JSON name fits to the HTML id, the value will be replaced
               "    if (document.getElementById('value'+i)) document.getElementById('value'+i).value = pwm2range(jo[i]);\n"    // in some cases we force to update the value, here for the PWM slider
               "    if (document.getElementById('cb'+i)) if(jo[i] == 1) document.getElementById('cb'+i).checked = true; else document.getElementById('cb'+i).checked = false;\n"
               "   }\n"  // update checkboxe
               // add other fields here (e.g. the delivered JSON name doesn't fit to the html id
               // finally, update the runtime
               "   if (jo['ss'] > 60) {document.getElementById('min').innerHTML = Math.floor(jo['ss'] / 60 % 60);}\n"
               "   if (jo['ss'] > 3600) {document.getElementById('hour').innerHTML = Math.floor(jo['ss'] / 3600 % 24);}\n"
               "   if (jo['ss'] > 86400) {document.getElementById('day').innerHTML = Math.floor(jo['ss'] / 86400 % 7);}\n"
               "   if (jo['ss'] > 604800) {document.getElementById('week').innerHTML = Math.floor(jo['ss'] / 604800 % 52);}\n"
               //"   document.getElementById('sec').style.color = 'dimgray';\n"                                     // if everything was ok, the second will be grey again.
               "   document.getElementById('sec').style.color = '';\n"                                              // if everything was ok, the second will be default color again
               "   if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&#10003;';\n"      // show activity: "OK"
               " })\n"
               " .catch(function() {\n"                                        // this is where you run code if the server returns any errors
               "  document.getElementById('sec').style.color = 'red';\n"
               " if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&#10005;';\n"        // show activity: "offline"
               " });\n"
               "}\n"
               "document.addEventListener('DOMContentLoaded', renew, setInterval(renew, ");
  message += ajaxIntervall * 1000;
  message += F("));\n");

  // smooth fading 64 values : https://www.mikrocontroller.net/articles/LED-Fading
  message += F("const pwmtable = [0, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, 7, 8, 9,\n"
               " 10, 11, 12, 13, 15, 17, 19, 21, 23, 26, 29, 32, 36, 40, 44, 49, 55,\n"
               " 61, 68, 76, 85, 94, 105, 117, 131, 146, 162, 181, 202, 225, 250,\n"
               " 279, 311, 346, 386, 430, 479, 534, 595, 663, 739, 824, 918, 1023];\n"
               // the smoothed fire function
               "function firesmooth(c,v){\n"      // c channel (mostly the id/self) | v myValue
               " v = pwmtable[v];\n"
               " var request = new XMLHttpRequest(); \n"
               " request.open('GET', 'cmd?range='+c+'&v='+v);\n"
               " request.send(null);\n"
               //" renew();\n"   // could be done, but fires very often
               "}\n"
               // a converter to get a from a pwmValue 0..1023 a range index 0..63
               "function pwm2range(v){\n"
               " for (i = 0; i < 64; i++){if (v <= pwmtable[i]) return i;}\n"
               " return 1023;\n"
               "}\n");

  message += F("function firecb(c){\n"                                 // c caller (mostly the id/self)
               " var request=new XMLHttpRequest();\n"
               " var v=0;\n"                                           // if value is initialisied with 0 we get ...
               " if (document.getElementById(c).checked) v=1;\n"       // ... 1 or 0 if checkbox is checked (better than true or false)
               " request.open('GET','cmd?cb='+c+'&v='+v,true);\n"
               " request.send(null);\n"
               "}\n");

  server.send(200, "text/javascript", message);
}
